// ---------------------------------------------------------------------------
//
//  This file is part of PermLib.
//
// Copyright (c) 2009-2011 Thomas Rehn <thomas@carmen76.de>
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote products
//    derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// ---------------------------------------------------------------------------

#ifndef DSET_H_
#define DSET_H_

#include <boost/dynamic_bitset.hpp>

namespace permlib {

/// a dynamic (bit)set
typedef boost::dynamic_bitset<> dset;

/// action of a permutation on a dynamic bitset
template<class PERM>
struct DSetAction {
	DSetAction(unsigned long size) : m_tmp(size) {}

	/// default action operator as also used in Transversal<PERM>::TrivialAction
	dset operator()(const PERM &p, const dset& v) {
		apply(p, v, m_tmp);
		return m_tmp;
	}

	// apply permutation in-place
	void apply(const PERM &p, const dset& v, dset& tmp) {
		tmp.reset();
		for (unsigned int i = 0; i < tmp.size(); ++i) {
			if (v[i])
				tmp.set(p / i, 1);
		}
	}
private:
	/// helper variable, allocated once
	dset m_tmp;
};


} // ns permlib
#endif // DSET_H_
